/*
 * Memaccess skin for Valgrind.
 * A thin callback layer for memory accesses.
 *
 * The instrumentation routine is based on the one in
 * Cachegrind (written by Nicholas Nethercote, njn25@cam.ac.uk)
 *
 * (C) 2003, Josef Weidendorfer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * The GNU General Public License is contained in the file COPYING.
 */

#include "vg_skin.h"
#include "memaccess.h"

VG_DETERMINE_INTERFACE_VERSION

/* According to IA-32 Intel Architecture Software Developer's Manual: Vol 2 */
#define MAX_x86_INSTR_SIZE 16
#define MIN_LINE_SIZE      16


/*--------------------------------------------------------------
 * Global skin state
 */

static ULong i_counter = 0;
static ULong r_counter = 0;
static ULong w_counter = 0;
static ULong i_counter_active = 0;
static ULong r_counter_active = 0;
static ULong w_counter_active = 0;

typedef void (*InstrReadProto)(Int tid, Addr iaddr, Int isize);
typedef void (*DataReadProto)(Int tid, Addr iaddr, Addr daddr, Int dsize);
typedef void (*DataWriteProto)(Int tid, Addr iaddr, Addr daddr, Int dsize);

static InstrReadProto ir_function = 0;
static DataReadProto dr_function = 0;
static DataWriteProto dw_function = 0;

static ThreadId current_tid = VG_INVALID_THREADID;
static ThreadId active_tid = VG_INVALID_THREADID;


/*--------------------------------------------------------------
 * Thread related
 *
 * <current_tid> keeps track of the current running thread.
 * If access notification is enabled, <active_tid> is equal to
 * <current_tid>. Otherwise, <active_tid> will be VG_INVALID_THREADID.
 */


static void thread_run(ThreadId tid)
{
    if (tid == current_tid) return;
    current_tid = tid;

    if (active_tid == VG_INVALID_THREADID) return;
    active_tid = tid;
}


/*--------------------------------------------------------------
 * Client Requests
 */

#ifndef VALGRIND_1_9_6
Bool SK_(handle_client_request)(ThreadId tid, UInt *args, UInt *ret)
#else
Bool SK_(handle_client_request)(ThreadState *tst, UInt *args, UInt *ret)
#endif
{
    static MAFuncTable ft;
    static int ft_inited = 0;

   if (!VG_IS_SKIN_USERREQ('M','A',args[0]))
      return False;

   switch(args[0]) {
   case MA_USERREQ__DEFINE_CALLBACKS:
       ir_function = (InstrReadProto) args[1];
       dr_function = (DataReadProto) args[2];
       dw_function = (DataWriteProto) args[3];

       if (!ft_inited) {
	   ft.printf = &(VG_(printf));
	   ft.malloc = &(VG_(malloc));
	   ft.free   = &(VG_(free));
	   ft.exit   = &(VG_(exit));
	   ft.panic  = &(VG_(skin_panic));
	   ft_inited = 1;
       }
       *ret = (UInt) &ft;

       if (VG_(clo_verbosity) > 2) {
	   VG_(message)(Vg_DebugMsg, 
			"Set Callbacks to (I %p, R %p, W %p",
			ir_function, dr_function, dw_function);
       }

       break;

   case MA_USERREQ__STOP:
       active_tid = VG_INVALID_THREADID;
       *ret = 0;

       if (VG_(clo_verbosity) > 2) {
	   VG_(message)(Vg_DebugMsg, "Stopped App Notification");
       }

       break;

   case MA_USERREQ__START:
       active_tid = current_tid;
       *ret = 0;

       if (VG_(clo_verbosity) > 2) {
	   VG_(message)(Vg_DebugMsg, "Started App Notification");
       }

       break;
 
   default:
       return False;
   }

   return True;
}



/*--------------------------------------------------------------
 * Callback related
 *--------------------------------------------------------------*/


static __attribute__ ((regparm (2)))
void access_I(Addr iaddr, int isize)
{
    i_counter++;
    if ((active_tid != VG_INVALID_THREADID) && ir_function) {

      if (VG_(clo_verbosity) > 2)
	VG_(message)(Vg_DebugMsg, "I Access (%p, size %d, tid %d) active",
		     iaddr, isize, active_tid);

	i_counter_active++;
	(*ir_function)(active_tid, iaddr, isize);
    }
    else if (VG_(clo_verbosity) > 3)
      VG_(message)(Vg_DebugMsg, "I Access (%p, size %d, tid %d)",
		   iaddr, isize, active_tid);
}

__attribute__ ((regparm (3))) static 
void access_R(Addr iaddr, Addr daddr, int dsize)
{
    r_counter++;
    if ((active_tid != VG_INVALID_THREADID) && dr_function) {

      if (VG_(clo_verbosity) > 2)
	VG_(message)(Vg_DebugMsg, "R Access (%p, size %d, tid %d) active",
		     daddr, dsize, active_tid);

	r_counter_active++;
	(*dr_function)(active_tid, iaddr, daddr, dsize);
    }
    else if (VG_(clo_verbosity) > 3)
      VG_(message)(Vg_DebugMsg, "R Access (%p, size %d, tid %d)",
		   daddr, dsize, active_tid);
}

__attribute__ ((regparm (3))) static 
void access_W(Addr iaddr, Addr daddr, int dsize)
{
    w_counter++;
    if ((active_tid != VG_INVALID_THREADID) && dw_function) {

    if (VG_(clo_verbosity) > 2)
	VG_(message)(Vg_DebugMsg, "W Access (%p, size %d, tid %d)",
		     daddr, dsize, active_tid);

	w_counter_active++;
	(*dw_function)(active_tid, iaddr, daddr, dsize);
    }
    else if (VG_(clo_verbosity) > 3)
      VG_(message)(Vg_DebugMsg, "W Access (%p, size %d, tid %d)",
		   daddr, dsize, active_tid);
}


/*--------------------------------------------------------------
 * Instrumentation
 *--------------------------------------------------------------*/

/**
 * This function is a highly modified version of the
 * instrument-function from cachegrind source
 */
UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
{
/* Use this rather than eg. -1 because it's a UInt. */
#define INVALID_DATA_SIZE   999999

   UCodeBlock* cb;
   Int         i;
   UInstr*     u_in;
   Int         t_instr_addr, t_instr_size, t_data_size;
   Int         t_read_addr, t_write_addr, t_read, t_write;
   Addr        x86_instr_addr = orig_addr;
   UInt        x86_instr_size = 1, data_size = INVALID_DATA_SIZE;
   Bool        instrumented_Jcond = False;
   Bool        has_rep_prefix     = False;

   cb = VG_(setup_UCodeBlock)(cb_in);

   t_instr_addr = t_read_addr = t_write_addr = INVALID_TEMPREG;
   t_instr_size = t_data_size = t_read = t_write = INVALID_TEMPREG;

   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++) {
      u_in = VG_(get_instr)(cb_in, i);

      /* What this is all about:  we want to instrument each x86 instruction 
       * translation.  The end of these are marked in three ways.  The three
       * ways, and the way we instrument them, are as follows:
       *
       * 1. UCode, INCEIP         --> UCode, Instrumentation, INCEIP
       * 2. UCode, Juncond        --> UCode, Instrumentation, Juncond
       * 3. UCode, Jcond, Juncond --> UCode, Instrumentation, Jcond, Juncond
       *
       * The last UInstr in a basic block is always a Juncond.  Jconds,
       * when they appear, are always second last.  We check this with 
       * various assertions.
       *
       * We must put the instrumentation before any jumps so that it is always
       * executed.  We don't have to put the instrumentation before the INCEIP
       * (it could go after) but we do so for consistency.
       *
       * x86 instruction sizes are obtained from INCEIPs (for case 1) or
       * from .extra4b field of the final JMP (for case 2 & 3).
       *
       * Note that JIFZ is treated differently.
       *
       * The instrumentation is just a call to the appropriate helper function,
       * passing it the address of the instruction's CC.
       */
      if (instrumented_Jcond) sk_assert(u_in->opcode == JMP);

      switch (u_in->opcode) {
         case NOP:  case LOCK:  case CALLM_E:  case CALLM_S:
            break;

         /* For memory-ref instrs, copy the data_addr into a temporary to be
          * passed to the cachesim_* helper at the end of the instruction.
          */
         case LOAD: 
            t_read      = u_in->val1;
            t_read_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val1,  TempReg, t_read_addr);
            data_size = u_in->size;
            VG_(copy_UInstr)(cb, u_in);
            break;

#ifndef VALGRIND_1_9_6
	    /* Newer Valgrind versions support MMX and SSE */

      case SSE2a_MemRd:
      case SSE2a1_MemRd:
            sk_assert(u_in->size == 4 || u_in->size == 16);
	    t_read = u_in->val3;
            t_read_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val3,  TempReg, t_read_addr);
            data_size = u_in->size;
            VG_(copy_UInstr)(cb, u_in);
            break;

      case SSE3a_MemRd:
            sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16);
	    t_read = u_in->val3;
            t_read_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val3,  TempReg, t_read_addr);
            data_size = u_in->size;
            VG_(copy_UInstr)(cb, u_in);
            break;

      case SSE3ag_MemRd_RegWr:
            sk_assert(u_in->size == 4 || u_in->size == 8);
	    t_read = u_in->val1;
            t_read_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val1,  TempReg, t_read_addr);
            data_size = u_in->size;
            VG_(copy_UInstr)(cb, u_in);
            break;

         case MMX2_MemRd:
            sk_assert(u_in->size == 4 || u_in->size == 8);
            /* fall through */
#endif
         case FPU_R:
            t_read      = u_in->val2;
            t_read_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val2,  TempReg, t_read_addr);
            data_size = ( u_in->size <= MIN_LINE_SIZE
                        ? u_in->size
                        : MIN_LINE_SIZE);
            VG_(copy_UInstr)(cb, u_in);
            break;

         /* Note that we must set t_write_addr even for mod instructions;
          * That's how the code above determines whether it does a write.
          * Without it, it would think a mod instruction is a read.
          * As for the MOV, if it's a mod instruction it's redundant, but it's
          * not expensive and mod instructions are rare anyway. */
#ifndef VALGRIND_1_9_6

      case SSE2a_MemWr:
            sk_assert(u_in->size == 4 || u_in->size == 16);
	    /* fall through */

      case SSE3a_MemWr:
           sk_assert(u_in->size == 4 || u_in->size == 8 || u_in->size == 16);
	    t_write = u_in->val3;
            t_write_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val3, TempReg, t_write_addr);
            data_size = u_in->size;
            VG_(copy_UInstr)(cb, u_in);
            break;

         case MMX2_MemWr:
            sk_assert(u_in->size == 4 || u_in->size == 8);
            /* fall through */
#endif

         case STORE:
         case FPU_W:
            t_write      = u_in->val2;
            t_write_addr = newTemp(cb);
            uInstr2(cb, MOV, 4, TempReg, u_in->val2, TempReg, t_write_addr);
            /* 28 and 108 B data-sized instructions will be done
             * inaccurately but they're very rare and this avoids errors
             * from hitting more than two cache lines in the simulation. */
            data_size = ( u_in->size <= MIN_LINE_SIZE
                        ? u_in->size
                        : MIN_LINE_SIZE);
            VG_(copy_UInstr)(cb, u_in);
            break;


         /* For rep-prefixed instructions, log a single I-cache access
          * before the UCode loop that implements the repeated part, which
          * is where the multiple D-cache accesses are logged. */
         case JIFZ:
            has_rep_prefix = True;

            /* instr addr */
            t_instr_addr = newTemp(cb);
            uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_addr);
            uLiteral(cb, x86_instr_addr);

	    /* JW - FIXME: instr_size is only given in INCEIP later ?! */

	    /* instr size */
            t_instr_size = newTemp(cb);
            uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_size);
            uLiteral(cb, x86_instr_size);

            /* Call helper */
            uInstr2(cb, CCALL, 0,
		    TempReg, t_instr_addr, 
		    TempReg, t_instr_size);
            uCCall(cb, (Addr) & access_I, 2, 2, False);
            VG_(copy_UInstr)(cb, u_in);
            break;


         /* INCEIP: insert instrumentation */
         case INCEIP:
            x86_instr_size = u_in->val1;
            goto instrument_x86_instr;

         /* JMP: insert instrumentation if the first JMP */
         case JMP:
            if (instrumented_Jcond) {
               sk_assert(CondAlways == u_in->cond);
               sk_assert(i+1 == VG_(get_num_instrs)(cb_in));
               VG_(copy_UInstr)(cb, u_in);
               instrumented_Jcond = False;    /* reset */
               break;
            }
            /* The first JMP... instrument. */
            if (CondAlways != u_in->cond) {
               sk_assert(i+2 == VG_(get_num_instrs)(cb_in));
               instrumented_Jcond = True;
            } else {
               sk_assert(i+1 == VG_(get_num_instrs)(cb_in));
            }

            /* Get x86 instr size from final JMP. */
            x86_instr_size = VG_(get_last_instr)(cb_in)->extra4b;

            goto instrument_x86_instr;


            /* Code executed at the end of each x86 instruction. */
            instrument_x86_instr:
             
            sk_assert(x86_instr_size >= 1 && 
                      x86_instr_size <= MAX_x86_INSTR_SIZE);

	    /* Call to access_I */
	    if (!has_rep_prefix) {

		/* instr addr */
		t_instr_addr = newTemp(cb);
		uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_addr);
		uLiteral(cb, x86_instr_addr);
		
		/* instr size */
		t_instr_size = newTemp(cb);
		uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_size);
		uLiteral(cb, x86_instr_size);

		/* Call helper */
		uInstr2(cb, CCALL, 0,
			TempReg, t_instr_addr, 
			TempReg, t_instr_size);
		uCCall(cb, (Addr) & access_I, 2, 2, False);
	    }


	    if ((t_read_addr != INVALID_TEMPREG) ||
		(t_write_addr != INVALID_TEMPREG)) {

		sk_assert(4 == data_size || 2  == data_size ||
			  1 == data_size || 
			  8 == data_size || 10 == data_size ||
			  MIN_LINE_SIZE == data_size);
	    }

	    if (t_read_addr != INVALID_TEMPREG) {

		if (t_instr_addr == INVALID_TEMPREG) {
		    /* instr addr */
		    t_instr_addr = newTemp(cb);
		    uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_addr);
		    uLiteral(cb, x86_instr_addr);
		}

		if (t_data_size == INVALID_TEMPREG) {
		    /* data size */
		    t_data_size = newTemp(cb);
		    uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_data_size);
		    uLiteral(cb, data_size);
		}

		/* Call helper */
		uInstr3(cb, CCALL, 0,
			TempReg, t_instr_addr,
			TempReg, t_read_addr, 
			TempReg, t_data_size);
		uCCall(cb, (Addr) & access_R, 3, 3, False);
	    }

	    if (t_write_addr != INVALID_TEMPREG) {

		if (t_instr_addr == INVALID_TEMPREG) {
		    /* instr addr */
		    t_instr_addr = newTemp(cb);
		    uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_instr_addr);
		    uLiteral(cb, x86_instr_addr);
		}

		if (t_data_size == INVALID_TEMPREG) {
		    /* data size */
		    t_data_size = newTemp(cb);
		    uInstr2(cb, MOV,  4, Literal, 0, TempReg, t_data_size);
		    uLiteral(cb, data_size);
		}

		/* Call helper */
		uInstr3(cb, CCALL, 0,
			TempReg, t_instr_addr,
			TempReg, t_write_addr, 
			TempReg, t_data_size);
		uCCall(cb, (Addr) & access_W, 3, 3, False);
	    }

            /* Copy original UInstr (INCEIP or JMP) */
            VG_(copy_UInstr)(cb, u_in);

            /* Update EIP, de-init read/write temps for next instr */
            x86_instr_addr += x86_instr_size;
            t_instr_addr = t_read_addr = t_write_addr = INVALID_TEMPREG;
	    t_instr_size = t_data_size = t_read = t_write = INVALID_TEMPREG;
            data_size = INVALID_DATA_SIZE;
            has_rep_prefix = False; 
            break;

         default:
            VG_(copy_UInstr)(cb, u_in);
            break;
      }
   }

   VG_(free_UCodeBlock)(cb_in);
   return cb;

#undef INVALID_DATA_SIZE
}


/*------------------------------------------------------------*/
/*--- SK_(fini)() and related function                     ---*/
/*------------------------------------------------------------*/

#ifdef VALGRIND_1_9_6
void SK_(fini)(void)
#else
void SK_(fini)(Int exitcode)
#endif
{
   VG_(message)(Vg_UserMsg, "I active/all: %llu / %llu", 
		i_counter_active, i_counter);
   VG_(message)(Vg_UserMsg, "R active/all: %llu / %llu", 
		r_counter_active, r_counter);
   VG_(message)(Vg_UserMsg, "W active/all: %llu / %llu", 
		w_counter_active, w_counter);
}


/*--------------------------------------------------------------------*/
/*--- Command line processing                                      ---*/
/*--------------------------------------------------------------------*/



/*--------------------------------------------------------------------*/
/*--- Setup                                                        ---*/
/*--------------------------------------------------------------------*/

void SK_(pre_clo_init)(void)
{
   VG_(details_name)            ("Memaccess");
   VG_(details_version)         (NULL);
   VG_(details_description)     ("a memory access trapper");
   VG_(details_copyright_author)(
      "Copyright (C) 2003, and GNU GPL'd, by Josef Weidendorfer.");
   VG_(details_bug_reports_to)  ("Josef.Weidendorfer@cs.tum.edu");
   VG_(details_avg_translation_sizeB) ( 155 );

   VG_(needs_client_requests)();
   VG_(track_thread_run) ( &thread_run );

   VG_(register_compact_helper)((Addr) & access_I);
   VG_(register_compact_helper)((Addr) & access_R);
   VG_(register_compact_helper)((Addr) & access_W);
}

void SK_(post_clo_init)(void)
{
}

